@@ -19,26 +19,69 @@ import (
 	"net"
 	"os"
 	"os/exec"
-	"path"
 	"path/filepath"
 	"regexp"
+	"runtime"
 	"strconv"
 	"strings"
 	"testing"
 	"time"
 
 	"github.com/kr/fs"
+	"github.com/stretchr/testify/assert"
 	"golang.org/x/crypto/ssh"
 )
 
-var testSftpClientBin = flag.String("sftp_client", "/usr/bin/sftp", "location of the sftp client binary")
+func TestMain(m *testing.M) {
+	sftpClientLocation, _ := exec.LookPath("sftp")
+	testSftpClientBin = flag.String("sftp_client", sftpClientLocation, "location of the sftp client binary")
+
+	lookSFTPServer := []string{
+		"/usr/libexec/sftp-server",
+		"/usr/lib/openssh/sftp-server",
+		"/usr/lib/ssh/sftp-server",
+		"C:\\Program Files\\Git\\usr\\lib\\ssh\\sftp-server.exe",
+	}
+	sftpServer, _ := exec.LookPath("sftp-server")
+	if len(sftpServer) == 0 {
+		for _, location := range lookSFTPServer {
+			if _, err := os.Stat(location); err == nil {
+				sftpServer = location
+				break
+			}
+		}
+	}
+	testSftp = flag.String("sftp", sftpServer, "location of the sftp server binary")
+	flag.Parse()
+
+	os.Exit(m.Run())
+}
+
+func skipIfWindows(t testing.TB) {
+	if runtime.GOOS == "windows" {
+		t.Skip("skipping test on windows")
+	}
+}
+
+func skipIfPlan9(t testing.TB) {
+	if runtime.GOOS == "plan9" {
+		t.Skip("skipping test on plan9")
+	}
+}
+
+var testServerImpl = flag.Bool("testserver", false, "perform integration tests against sftp package server instance")
+var testIntegration = flag.Bool("integration", false, "perform integration tests against sftp server process")
+var testAllocator = flag.Bool("allocator", false, "perform tests using the allocator")
+var testSftp *string
+
+var testSftpClientBin *string
 var sshServerDebugStream = ioutil.Discard
 var sftpServerDebugStream = ioutil.Discard
 var sftpClientDebugStream = ioutil.Discard
 
 const (
-	GOLANG_SFTP  = true
-	OPENSSH_SFTP = false
+	GolangSFTP  = true
+	OpenSSHSFTP = false
 )
 
 var (
@@ -319,7 +362,9 @@ func (chsvr *sshSessionChannelServer) handleSubsystem(req *ssh.Request) error {
 }
 
 // starts an ssh server to test. returns: host string and port
-func testServer(t *testing.T, useSubsystem bool, readonly bool) (net.Listener, string, int) {
+func testServer(t *testing.T, useSubsystem bool, readonly bool) (func(), string, int) {
+	t.Helper()
+
 	if !*testIntegration {
 		t.Skip("skipping integration test")
 	}
@@ -338,43 +383,50 @@ func testServer(t *testing.T, useSubsystem bool, readonly bool) (net.Listener, s
 		t.Fatal(err)
 	}
 
+	shutdown := make(chan struct{})
+
 	go func() {
 		for {
 			conn, err := listener.Accept()
 			if err != nil {
-				fmt.Fprintf(sshServerDebugStream, "ssh server socket closed: %v\n", err)
-				break
+				select {
+				case <-shutdown:
+				default:
+					t.Error("ssh server socket closed:", err)
+				}
+				return
 			}
 
 			go func() {
 				defer conn.Close()
+
 				sshSvr, err := sshServerFromConn(conn, useSubsystem, basicServerConfig())
 				if err != nil {
 					t.Error(err)
 					return
 				}
-				err = sshSvr.Wait()
-				fmt.Fprintf(sshServerDebugStream, "ssh server finished, err: %v\n", err)
+
+				_ = sshSvr.Wait()
 			}()
 		}
 	}()
 
-	return listener, host, port
+	return func() { close(shutdown); listener.Close() }, host, port
 }
 
 func makeDummyKey() (string, error) {
 	priv, err := ecdsa.GenerateKey(elliptic.P256(), crand.Reader)
 	if err != nil {
-		return "", fmt.Errorf("cannot generate key: %v", err)
+		return "", fmt.Errorf("cannot generate key: %w", err)
 	}
 	der, err := x509.MarshalECPrivateKey(priv)
 	if err != nil {
-		return "", fmt.Errorf("cannot marshal key: %v", err)
+		return "", fmt.Errorf("cannot marshal key: %w", err)
 	}
 	block := &pem.Block{Type: "EC PRIVATE KEY", Bytes: der}
 	f, err := ioutil.TempFile("", "sftp-test-key-")
 	if err != nil {
-		return "", fmt.Errorf("cannot create temp file: %v", err)
+		return "", fmt.Errorf("cannot create temp file: %w", err)
 	}
 	defer func() {
 		if f != nil {
@@ -383,16 +435,34 @@ func makeDummyKey() (string, error) {
 		}
 	}()
 	if err := pem.Encode(f, block); err != nil {
-		return "", fmt.Errorf("cannot write key: %v", err)
+		return "", fmt.Errorf("cannot write key: %w", err)
 	}
 	if err := f.Close(); err != nil {
-		return "", fmt.Errorf("error closing key file: %v", err)
+		return "", fmt.Errorf("error closing key file: %w", err)
 	}
 	path := f.Name()
 	f = nil
 	return path, nil
 }
 
+type execError struct {
+	path   string
+	stderr string
+	err    error
+}
+
+func (e *execError) Error() string {
+	return fmt.Sprintf("%s: %v: %s", e.path, e.err, e.stderr)
+}
+
+func (e *execError) Unwrap() error {
+	return e.err
+}
+
+func (e *execError) Cause() error {
+	return e.err
+}
+
 func runSftpClient(t *testing.T, script string, path string, host string, port int) (string, error) {
 	// if sftp client binary is unavailable, skip test
 	if _, err := os.Stat(*testSftpClientBin); err != nil {
@@ -406,34 +476,85 @@ func runSftpClient(t *testing.T, script string, path string, host string, port i
 	}
 	defer os.Remove(dummyKey)
 
-	args := []string{
+	cmd := exec.Command(
+		*testSftpClientBin,
 		// "-vvvv",
 		"-b", "-",
 		"-o", "StrictHostKeyChecking=no",
 		"-o", "LogLevel=ERROR",
 		"-o", "UserKnownHostsFile /dev/null",
 		// do not trigger ssh-agent prompting
-		"-o", "IdentityFile=" + dummyKey,
+		"-o", "IdentityFile="+dummyKey,
 		"-o", "IdentitiesOnly=yes",
 		"-P", fmt.Sprintf("%d", port), fmt.Sprintf("%s:%s", host, path),
-	}
-	cmd := exec.Command(*testSftpClientBin, args...)
-	var stdout bytes.Buffer
-	cmd.Stdin = bytes.NewBufferString(script)
-	cmd.Stdout = &stdout
-	cmd.Stderr = sftpClientDebugStream
+	)
+
+	cmd.Stdin = strings.NewReader(script)
+
+	stdout := new(bytes.Buffer)
+	cmd.Stdout = stdout
+
+	stderr := new(bytes.Buffer)
+	cmd.Stderr = stderr
+
 	if err := cmd.Start(); err != nil {
 		return "", err
 	}
-	err = cmd.Wait()
-	return stdout.String(), err
+
+	if err := cmd.Wait(); err != nil {
+		return stdout.String(), &execError{
+			path:   cmd.Path,
+			stderr: stderr.String(),
+			err:    err,
+		}
+	}
+
+	return stdout.String(), nil
+}
+
+// assert.Eventually seems to have a data rate on macOS with go 1.14 so replace it with this simpler function
+func waitForCondition(t *testing.T, condition func() bool) {
+	start := time.Now()
+	tick := 10 * time.Millisecond
+	waitFor := 100 * time.Millisecond
+	for !condition() {
+		time.Sleep(tick)
+		if time.Since(start) > waitFor {
+			break
+		}
+	}
+	assert.True(t, condition())
+}
+
+func checkAllocatorBeforeServerClose(t *testing.T, alloc *allocator) {
+	if alloc != nil {
+		// before closing the server we are, generally, waiting for new packets in recvPacket and we have a page allocated.
+		// Sometime the sendPacket returns some milliseconds after the client receives the response, and so we have 2
+		// allocated pages here, so wait some milliseconds. To avoid crashes we must be sure to not release the pages
+		// too soon.
+		waitForCondition(t, func() bool { return alloc.countUsedPages() <= 1 })
+	}
+}
+
+func checkAllocatorAfterServerClose(t *testing.T, alloc *allocator) {
+	if alloc != nil {
+		// wait for the server cleanup
+		waitForCondition(t, func() bool { return alloc.countUsedPages() == 0 })
+		waitForCondition(t, func() bool { return alloc.countAvailablePages() == 0 })
+	}
 }
 
 func TestServerCompareSubsystems(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	listenerOp, hostOp, portOp := testServer(t, OPENSSH_SFTP, READONLY)
-	defer listenerGo.Close()
-	defer listenerOp.Close()
+	if runtime.GOOS == "windows" {
+		// TODO (puellanivis): not sure how to fix this, the OpenSSH SFTP implementation closes immediately.
+		t.Skip()
+	}
+
+	shutdownGo, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdownGo()
+
+	shutdownOp, hostOp, portOp := testServer(t, OpenSSHSFTP, READONLY)
+	defer shutdownOp()
 
 	script := `
 ls /
@@ -459,10 +580,11 @@ ls -l /usr/bin/
 	outputGoLines := newlineRegex.Split(outputGo, -1)
 	outputOpLines := newlineRegex.Split(outputOp, -1)
 
+	if len(outputGoLines) != len(outputOpLines) {
+		t.Fatalf("output line count differs, go = %d, openssh = %d", len(outputGoLines), len(outputOpLines))
+	}
+
 	for i, goLine := range outputGoLines {
-		if i > len(outputOpLines) {
-			t.Fatalf("output line count differs")
-		}
 		opLine := outputOpLines[i]
 		bad := false
 		if goLine != opLine {
@@ -474,8 +596,9 @@ ls -l /usr/bin/
 			// during testing as processes are created/destroyed.
 			// words[7] as timestamp on dirs can very for things like /tmp
 			for j, goWord := range goWords {
-				if j > len(opWords) {
+				if j >= len(opWords) {
 					bad = true
+					break
 				}
 				opWord := opWords[j]
 				if goWord != opWord && j != 1 && j != 2 && j != 3 && j != 7 {
@@ -485,7 +608,7 @@ ls -l /usr/bin/
 		}
 
 		if bad {
-			t.Errorf("outputs differ, go:\n%v\nopenssh:\n%v\n", goLine, opLine)
+			t.Errorf("outputs differ\n     go: %q\nopenssh: %q\n", goLine, opLine)
 		}
 	}
 }
@@ -505,8 +628,8 @@ func randName() string {
 }
 
 func TestServerMkdirRmdir(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	defer listenerGo.Close()
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
 
 	tmpDir := "/tmp/" + randName()
 	defer os.RemoveAll(tmpDir)
@@ -531,9 +654,39 @@ func TestServerMkdirRmdir(t *testing.T) {
 	}
 }
 
+func TestServerLink(t *testing.T) {
+	skipIfWindows(t) // No hard links on windows.
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
+
+	tmpFileLocalData := randData(999)
+
+	linkdest := "/tmp/" + randName()
+	defer os.RemoveAll(linkdest)
+	if err := ioutil.WriteFile(linkdest, tmpFileLocalData, 0644); err != nil {
+		t.Fatal(err)
+	}
+
+	link := "/tmp/" + randName()
+	defer os.RemoveAll(link)
+
+	// now create a hard link within the new directory
+	if output, err := runSftpClient(t, fmt.Sprintf("ln %s %s", linkdest, link), "/", hostGo, portGo); err != nil {
+		t.Fatalf("failed: %v %v", err, string(output))
+	}
+
+	// file should now exist and be the same size as linkdest
+	if stat, err := os.Lstat(link); err != nil {
+		t.Fatal(err)
+	} else if int(stat.Size()) != len(tmpFileLocalData) {
+		t.Fatalf("wrong size: %v", len(tmpFileLocalData))
+	}
+}
+
 func TestServerSymlink(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	defer listenerGo.Close()
+	skipIfWindows(t) // No symlinks on windows.
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
 
 	link := "/tmp/" + randName()
 	defer os.RemoveAll(link)
@@ -552,8 +705,8 @@ func TestServerSymlink(t *testing.T) {
 }
 
 func TestServerPut(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	defer listenerGo.Close()
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
 
 	tmpFileLocal := "/tmp/" + randName()
 	tmpFileRemote := "/tmp/" + randName()
@@ -581,9 +734,56 @@ func TestServerPut(t *testing.T) {
 	}
 }
 
+func TestServerResume(t *testing.T) {
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
+
+	tmpFileLocal := "/tmp/" + randName()
+	tmpFileRemote := "/tmp/" + randName()
+	defer os.RemoveAll(tmpFileLocal)
+	defer os.RemoveAll(tmpFileRemote)
+
+	t.Logf("put: local %v remote %v", tmpFileLocal, tmpFileRemote)
+
+	// create a local file with random contents to be pushed to the server
+	tmpFileLocalData := randData(2 * 1024 * 1024)
+	// only write half the data to simulate a split upload
+	half := 1024 * 1024
+	err := ioutil.WriteFile(tmpFileLocal, tmpFileLocalData[:half], 0644)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// sftp the first half of the file to the server
+	output, err := runSftpClient(t, "put "+tmpFileLocal+" "+tmpFileRemote,
+		"/", hostGo, portGo)
+	if err != nil {
+		t.Fatalf("runSftpClient failed: %v, output\n%v\n", err, output)
+	}
+
+	// write the full file out
+	err = ioutil.WriteFile(tmpFileLocal, tmpFileLocalData, 0644)
+	if err != nil {
+		t.Fatal(err)
+	}
+	// re-sftp the full file with the append flag set
+	output, err = runSftpClient(t, "put -a "+tmpFileLocal+" "+tmpFileRemote,
+		"/", hostGo, portGo)
+	if err != nil {
+		t.Fatalf("runSftpClient failed: %v, output\n%v\n", err, output)
+	}
+
+	// tmpFileRemote should now exist, with the same contents
+	if tmpFileRemoteData, err := ioutil.ReadFile(tmpFileRemote); err != nil {
+		t.Fatal(err)
+	} else if string(tmpFileLocalData) != string(tmpFileRemoteData) {
+		t.Fatal("contents of file incorrect after put")
+	}
+}
+
 func TestServerGet(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	defer listenerGo.Close()
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
 
 	tmpFileLocal := "/tmp/" + randName()
 	tmpFileRemote := "/tmp/" + randName()
@@ -623,7 +823,7 @@ func compareDirectoriesRecursive(t *testing.T, aroot, broot string) {
 		if err != nil {
 			t.Fatalf("could not find relative path for %v: %v", aPath, err)
 		}
-		bPath := path.Join(broot, aRel)
+		bPath := filepath.Join(broot, aRel)
 
 		if aRel == "." {
 			continue
@@ -678,8 +878,8 @@ func compareDirectoriesRecursive(t *testing.T, aroot, broot string) {
 }
 
 func TestServerPutRecursive(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	defer listenerGo.Close()
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
 
 	dirLocal, err := os.Getwd()
 	if err != nil {
@@ -690,17 +890,23 @@ func TestServerPutRecursive(t *testing.T) {
 
 	t.Logf("put recursive: local %v remote %v", dirLocal, tmpDirRemote)
 
+	// On windows, the client copies the contents of the directory, not the directory itself.
+	winFix := ""
+	if runtime.GOOS == "windows" {
+		winFix = "/" + filepath.Base(dirLocal)
+	} //*/
+
 	// push this directory (source code etc) recursively to the server
-	if output, err := runSftpClient(t, "mkdir "+tmpDirRemote+"\r\nput -r -P "+dirLocal+"/ "+tmpDirRemote+"/", "/", hostGo, portGo); err != nil {
+	if output, err := runSftpClient(t, "mkdir "+tmpDirRemote+"\r\nput -R -p "+dirLocal+" "+tmpDirRemote+winFix, "/", hostGo, portGo); err != nil {
 		t.Fatalf("runSftpClient failed: %v, output\n%v\n", err, output)
 	}
 
-	compareDirectoriesRecursive(t, dirLocal, path.Join(tmpDirRemote, path.Base(dirLocal)))
+	compareDirectoriesRecursive(t, dirLocal, filepath.Join(tmpDirRemote, filepath.Base(dirLocal)))
 }
 
 func TestServerGetRecursive(t *testing.T) {
-	listenerGo, hostGo, portGo := testServer(t, GOLANG_SFTP, READONLY)
-	defer listenerGo.Close()
+	shutdown, hostGo, portGo := testServer(t, GolangSFTP, READONLY)
+	defer shutdown()
 
 	dirRemote, err := os.Getwd()
 	if err != nil {
@@ -711,10 +917,16 @@ func TestServerGetRecursive(t *testing.T) {
 
 	t.Logf("get recursive: local %v remote %v", tmpDirLocal, dirRemote)
 
+	// On windows, the client copies the contents of the directory, not the directory itself.
+	winFix := ""
+	if runtime.GOOS == "windows" {
+		winFix = "/" + filepath.Base(dirRemote)
+	}
+
 	// pull this directory (source code etc) recursively from the server
-	if output, err := runSftpClient(t, "lmkdir "+tmpDirLocal+"\r\nget -r -P "+dirRemote+"/ "+tmpDirLocal+"/", "/", hostGo, portGo); err != nil {
+	if output, err := runSftpClient(t, "lmkdir "+tmpDirLocal+"\r\nget -R -p "+dirRemote+" "+tmpDirLocal+winFix, "/", hostGo, portGo); err != nil {
 		t.Fatalf("runSftpClient failed: %v, output\n%v\n", err, output)
 	}
 
-	compareDirectoriesRecursive(t, dirRemote, path.Join(tmpDirLocal, path.Base(dirRemote)))
+	compareDirectoriesRecursive(t, dirRemote, filepath.Join(tmpDirLocal, filepath.Base(dirRemote)))
 }
